% main function for the FRAME model for general images
clear all;
close all;

nOrient = 16;
localHalfx=20;
localHalfy=20;
thresholdFactor=0.01;
nIteration=500;

isLocalNormalize=false;
isV2=false;  % it matters, if isLocalNormalize=true;  %% V1 is Wenze's version, V2 is Zhangzhang's version
isSeparate=false;  % it matters, if isV2=false;

isToGenerateMovie=false;
isFreeBoundary=true;
useDoG=true;

sx=60;
sy=60;

lambdaLearningRate = 0.01;
epsilon = 0.003;
L= 30;


scaleList=[0.4, 0.8, 0.2]; % you can add more scales of Gabor filters here

%% Step 0: prepare filter and training images
mex -largeArrayDims CgetMAX1.c;
%mex mexc_integralMap.cpp;
mex Ctransform.c;
%mex -largeArrayDims mexc_GibbsSampling.cpp;
%mex CgetMAX1.c;
%mex mexc_GibbsSampling.cpp;

nScaleGabor=length(scaleList);

filters=[];
for iScale=1:nScaleGabor

    f = MakeFilter(scaleList(iScale),nOrient);
    for i=1:nOrient
     f_r{i} =real(f{i});
     f_i{i} =imag(f{i});
    end
    
    filters = [filters f_r f_i];
 
end

if useDoG
 f0 = dog(10,0);
 filters = [filters f0];
end

numFilter = length(filters);

halfFilterSize = zeros(size(filters));
for iF = 1:numFilter
   halfFilterSize(iF)=(size(filters{iF},1)-1)/2;
end


locationShiftLimit=0;
orientShiftLimit=0;

inPath = './positiveImage';
imgCell = cell(0);
files = dir([inPath '/*.jpg']);
for iImg = 1:length(files)
    img = imread(fullfile(inPath,files(iImg).name));
    img = imresize(img,[sx, sy]);
    im=im2double(rgb2gray(img));
    im=im-mean(im(:));
    im=im/std(im(:));
    imgCell{iImg}=im;
end
%% Step 1: compute training sample averages

img = rgb2gray(img);
rHat = cell(numFilter,1);
for iFilter = 1:numFilter
    rHat{iFilter}=zeros(sx,sy);
end
tic;
for iImg = 1:length(files)
    S1 = cell(numFilter,1);
    M1 = cell(numFilter,1);
    
    % SUM1
    for iFilter = 1:numFilter
                
        Y = filter2(filters{iFilter},imgCell{iImg});
        S1{iFilter} = abs(single(Y));
        M1{iFilter} = zeros(sx,sy,'single');
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Local Normalization   
    if isLocalNormalize
    
    if isV2  % V2 use Separate local normalization
      for iScale=1:nScaleGabor 
        h = halfFilterSize(1+(iScale-1)*(2*nOrient));
        S1(1+(iScale-1)*(2*nOrient) : nOrient+(iScale-1)*(2*nOrient))= LocalNormalizeV2(S1(1+(iScale-1)*(2*nOrient):nOrient+(iScale-1)*(2*nOrient)),h,localHalfx,localHalfy,thresholdFactor);
        S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient)) = LocalNormalizeV2(S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient)),h,localHalfx,localHalfy,thresholdFactor);
      end
      if useDoG
         h = halfFilterSize(end);       
         S1(end)=LocalNormalizeV2(S1(end),h,localHalfx,localHalfy,thresholdFactor);                
      end
    else
    
    %%%%% V1 
    if isSeparate        
      for iScale=1:nScaleGabor  
        h = halfFilterSize(1+(iScale-1)*(2*nOrient)); 
        [S1(1+(iScale-1)*(2*nOrient) : nOrient+(iScale-1)*(2*nOrient))]= LocalNormalize(S1(1+(iScale-1)*(2*nOrient):nOrient+(iScale-1)*(2*nOrient)),[],h,localHalfx,localHalfy,thresholdFactor);
        [S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient))] = LocalNormalize(S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient)),[],h,localHalfx,localHalfy,thresholdFactor);
      end
      if useDoG
         h = halfFilterSize(end);       
        [S1(end)]=LocalNormalize(S1(end),[],h,localHalfx,localHalfy,thresholdFactor);                
      end    
    else        
        for iScale=1:nScaleGabor
           h = halfFilterSize(1+(iScale-1)*(2*nOrient)); 
           [S1(1+(iScale-1)*(2*nOrient):nOrient+(iScale-1)*(2*nOrient)),S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient))]= LocalNormalize(S1(1+(iScale-1)*(2*nOrient):nOrient+(iScale-1)*(2*nOrient)),S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient)),h,localHalfx,localHalfy,thresholdFactor);
        end
        if useDoG
           h = halfFilterSize(end);       
           [S1(end)]=LocalNormalize(S1(end),[],h,localHalfx,localHalfy,thresholdFactor);                
        end        
    end
    %%%%%% end of V1
    
    end
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
    %MAX1
    for iScale=1:nScaleGabor
      CgetMAX1(1,sx,sy,nOrient,locationShiftLimit,orientShiftLimit,1,S1(1+(iScale-1)*(2*nOrient):nOrient+(iScale-1)*(2*nOrient)),M1(1+(iScale-1)*(2*nOrient):nOrient+(iScale-1)*(2*nOrient)));
      CgetMAX1(1,sx,sy,nOrient,locationShiftLimit,orientShiftLimit,1,S1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient)),M1(nOrient+1+(iScale-1)*(2*nOrient): 2*nOrient+(iScale-1)*(2*nOrient)));
    end
    if useDoG 
      M1{end}=S1{end};
    end
    
    for iFilter = 1:numFilter
        rHat{iFilter}= rHat{iFilter}+double(M1{iFilter});
    end
end
for iFilter = 1:numFilter
    rHat{iFilter}=rHat{iFilter}/length(files);
end
disp(['finished filtering: '  num2str(toc) ' seconds']);


%% Step 2: optimize the exponential model
% our model parameter
lambdaF = cell(numFilter,1);
gradientF = cell(numFilter,1);
for iFilter = 1:numFilter
    lambdaF{iFilter} = zeros(size(img));%rand(size(img));
    gradientF{iFilter}= zeros(size(img));
end

currSample = randn(size(img));
% store the <F, I> of current sample
rSample = cell(numFilter,1);
for iFilter = 1:numFilter
    rSample{iFilter}=abs(filter2(filters{iFilter},currSample));
end

SSD=zeros(nIteration,1);
nValidFeature= sum((sx-2*halfFilterSize).*(sy-2*halfFilterSize));

for iter = 1:nIteration
    disp( [ 'iteration: ' num2str(iter)]);
    tic 
        
    currSample=HMC_I( numFilter,lambdaF, filters, currSample, epsilon, L);
    for iFilter = 1:numFilter
      rModel{iFilter}=abs(filter2(filters{iFilter},currSample));
    end
    
    disp(['one iteration learning time: ' num2str(toc) ' seconds']);

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5   
    % compute gradient and do graidnet ascent
           
    gradientNorm =0;
    
    for iFilter = 1:numFilter
        gradientF{iFilter} = rHat{iFilter}-rModel{iFilter};
        h = halfFilterSize(iFilter);
        if isFreeBoundary
          gradientF{iFilter}([1:h,sx-h+1:sx],:)=0;
          gradientF{iFilter}(:,[1:h,sy-h+1:sy])=0;
        end
        aa=gradientF{iFilter};          
        gradientNorm = gradientNorm + sum(sum(abs(aa(h+1:end-h,h+1:end-h))));
    end
    
    SSD(iter)=gradientNorm/nValidFeature;
    
    for iFilter = 1:numFilter
        lambdaF{iFilter}=lambdaF{iFilter}+ lambdaLearningRate * gradientF{iFilter};
    end
   
    % save synthesied image
    gLow = min(currSample(:));
    gHigh = max(currSample(:));
    disp([ 'min: ' num2str(gLow) ' max: ' num2str(gHigh)]);
    disp([ 'SSD: ' num2str(SSD(iter))]);
    img = (currSample-gLow)/(gHigh-gLow);
    imwrite(img,[ num2str(iter,'%04d') '.png']);
    %imwrite(1-img,[ num2str(iter,'%04d') '.png']);
end

if isToGenerateMovie
filename='movie.gif';
for i=1:nIteration
    str=[ num2str(i,'%04d') '.png'];
    Img=imread(str);
    figure(i)
    imshow(Img,'border','tight','initialmagnification','fit');
    set(gcf,'Position',[50,50,130,130]);     % the dimension of the gif should be lareger than this default setting.
    axis normal;
    frame=getframe(i);
    im=frame2im(frame);
    [I,map]=rgb2ind(im,256);
    k=i-0;
    close;
    if k==1;
        imwrite(I,map,filename,'gif','Loopcount',inf,'DelayTime',0.15);
    else
        imwrite(I,map,filename,'gif','WriteMode','append','DelayTime',0.15);
    end
end
end

h = figure; 
plot(1:nIteration,SSD);
title('Convergence');
xlabel('Iterations');
ylabel('SSD');
saveas(h,'plot.bmp','bmp');


